import React, { useContext, useEffect, useState } from 'react';
import PropTypes from 'prop-types';
import { connect } from 'react-redux';
import moment from 'moment';
import interceptClient from 'interceptClient';
import find from 'lodash/find';
import get from 'lodash/get';


// Local Components
import Calendar from './ReservationCalendar';
import RoomsContext from '../context/RoomsContext';
import GroupsContext from '../context/GroupsContext';
import useEventListener from '../hooks/useEventListener';

const { constants, api, select, utils } = interceptClient;
const c = constants;

const ADD_ROOM_RESERVATION = 'intercept:addRoomReservation';
const EDIT_ROOM_RESERVATION = 'intercept:editRoomReservation';
const VIEW_ROOM_RESERVATION = 'intercept:viewRoomReservation';
const SAVE_ROOM_RESERVATION = 'intercept:saveRoomReservation';
const SAVE_ROOM_RESERVATION_SUCCESS = 'intercept:saveRoomReservationSuccess';
const SAVE_ROOM_RESERVATION_ERROR = 'intercept:saveRoomReservationError';
const CHANGE_ROOM_RESERVATION = 'intercept:changeRoomReservation';
const CLOSE_ROOM_RESERVATION = 'intercept:closeRoomReservation';
const REFRESH_ROOM_RESERVATION = 'intercept:updateRoomReservation';

const getResourceGroupFromLocation = (location, getHours) => ({
  id: location.data.id,
  title: location.data.attributes.title,
  hours: getHours(location.data.id),
});

const getResourceGroupsFromLocations = (locations, getHours) =>
  locations.map(location => getResourceGroupFromLocation(location, getHours));

const getResourceFromRoom = (room, getHours) => ({
  id: room.id,
  title: room.attributes.title,
  groupId: room.relationships.field_location.data[0].id,
  hours: getHours(room.relationships.field_location.data[0].id),
  drupal_internal__nid: get(room, 'attributes.drupal_internal__nid'),
});

const getResourcesFromRooms = (rooms, getHours) =>
  rooms.map(room => getResourceFromRoom(room, getHours));

const getEventFromReservation = reservation => ({
  id: get(reservation, 'data.id'),
  title: get(reservation, 'data.attributes.field_group_name') || 'Booked',
  resourceId: get(reservation, 'data.relationships.field_room.data.id'),
  status: get(reservation, 'data.attributes.field_status'),
  start: get(reservation, 'data.attributes.field_dates.value'),
  end: get(reservation, 'data.attributes.field_dates.end_value'),
  drupal_internal__id: get(reservation, 'data.attributes.drupal_internal__id')
});

const getEventsFromReservations = reservation => reservation.map(getEventFromReservation);

const reservationHasRoom = reservation => reservation.data.relationships.field_room.data.id !== 'missing';

/**
 * Creates an Intercept ADD reservation event to be dispatched
 *
 * @param {Object} reservation
 *   Room reservation id.
 * @param {[Object]} resources
 *   An array of available resources.
 * @return {CustomEvent}
 *   A Custom Event object.
 */
function getAddRoomReservationEvent(reservation, resources) {
  const tz = utils.getUserTimezone();
  const start = moment(reservation.start).tz(tz);
  const end = moment(reservation.end).tz(tz);
  const resource = resources.find(item => item.id === reservation.resourceId);

  const event = new CustomEvent(ADD_ROOM_RESERVATION, {
    detail: {
      ...reservation.event,
      resource,
      end: {
        date: end.format('YYYY-MM-DD'),
        time: end.format('HH:mm:ss'),
      },
      start: {
        date: start.format('YYYY-MM-DD'),
        time: start.format('HH:mm:ss'),
      },
    },
  });

  return event;
}

/**
 * Creates an Intercept CHANGE reservation event to be dispatched
 *
 * @param {Object} reservation
 *   Room reservation id.
 * @param {[Object]} resources
 *   An array of available resources.
 * @return {CustomEvent}
 *   A Custom Event object.
 */
function getChangeRoomReservationEvent(reservation, resources) {
  const tz = utils.getUserTimezone();
  const start = moment(reservation.start).tz(tz);
  const end = moment(reservation.end).tz(tz);
  const resource = resources.find(item => item.id === reservation.resourceId);

  const event = new CustomEvent(CHANGE_ROOM_RESERVATION, {
    detail: {
      ...reservation.event,
      resource,
      end: {
        date: end.format('YYYY-MM-DD'),
        time: end.format('HH:mm:ss'),
      },
      start: {
        date: start.format('YYYY-MM-DD'),
        time: start.format('HH:mm:ss'),
      },
    },
  });

  return event;
}

/**
 * Creates an Intercept VIEW reservation event to be dispatched
 *
 * @param {string} id
 *   Room reservation id.
 * @return {CustomEvent}
 *   A Custom Event object.
 */
function getSelectRoomReservationEvent(id) {
  const event = new CustomEvent(VIEW_ROOM_RESERVATION, {
    detail: {
      id,
    },
  });

  return event;
}

/**
 * Creates an Intercept EDIT reservation event to be dispatched
 *
 * @param {string} id
 *   Room reservation id.
 * @return {CustomEvent}
 *   A Custom Event object.
 */
function getEditRoomReservationEvent(id) {
  const event = new CustomEvent(EDIT_ROOM_RESERVATION, {
    detail: {
      id,
    },
  });

  return event;
}

function getDate(value, view = 'day', boundary = 'start') {
  const method = boundary === 'start' ? 'startOf' : 'endOf';
  const date = moment.tz(value, utils.getUserTimezone())[method](view);
  // The calendar view may include date from the previous or next month
  // so we make sure to include the beginning of the first week and
  // end of the last week.
  if (view === 'month') {
    date[method]('week');
  }

  return date.format();
}

function getDateSpan(value, view = 'day') {
  const start = getDate(value, view, 'start');
  const end = getDate(value, view, 'end');
  return [start, end];
}

function getDateFilters(date = new Date(), view = 'day') {
  const path = 'field_dates.value';
  // Handler Calendar view.
  // The date should be determined by the date and calendar view type
  // rather than the selected date value.
  const value = getDateSpan(date, view);
  const operator = 'BETWEEN';

  return {
    data: {
      path,
      value,
      operator,
    },
  };
}

function getRoomFilters(rooms = []) {
  return {
    location: {
      path: 'field_room.id',
      value: rooms,
      operator: 'IN',
    },
  };
}

const RoomReservationScheduler = ({
  date,
  getLocationHours,
  fetchLocations,
  fetchReservations,
  onChangeDate,
  onChangeView,
  purgeReservations,
  locations,
  reservations,
  view,
}) => {
  const { rooms } = useContext(RoomsContext);
  const { setGroups } = useContext(GroupsContext);

  const doFetchReservations = () => {
    fetchReservations({
      filters: {
        ...getDateFilters(date, view),
        ...getRoomFilters(rooms.map(room => room.id)),
        status: {
          path: 'field_status',
          value: ['approved', 'requested'],
          operator: 'IN',
        },
      },
      fields: {
        [c.TYPE_ROOM_RESERVATION]: ['field_group_name', 'field_room', 'field_dates', 'field_status', 'drupal_internal__id'],
      },
      replace: true,
      headers: {
        'X-Consumer-ID': interceptClient.consumer,
      },
    });
  };

  const [
    selectedEvent,
    setSelectedEvent,
  ] = useState(null);

  // Component Did Mount
  useEffect(() => {
    fetchLocations({
      filters: {
        onlyBranchLocation: {
          path: 'field_branch_location',
          value: '1',
        },
        published: {
          path: 'status',
          value: '1',
        },
      },
      fields: {
        [c.TYPE_LOCATION]: ['title', 'field_location_hours', 'field_branch_location'],
      },
    });
  }, []);

  useEffect(() => {
    // Avoid overloading the system by only fetching if we are filtering by rooms.
    if (rooms.length > 0) {
      doFetchReservations();
    }
  }, [date, view, rooms]);

  useEffect(() => {
    setGroups(getResourceGroupsFromLocations(locations, getLocationHours));
  }, [locations]);

  const resources = getResourcesFromRooms(rooms, getLocationHours);
  const events = getEventsFromReservations(reservations);

  const onDoubleClickEvent = (event) => {
    window.dispatchEvent(getEditRoomReservationEvent(event.drupal_internal__id));
  };

  const onChangeEvent = (event) => {
    setSelectedEvent({
      ...event.event,
      start: moment(event.start).tz(utils.getUserTimezone()).format(),
      end: moment(event.end).tz(utils.getUserTimezone()).format(),
      resourceId: event.resourceId,
    });
    window.dispatchEvent(getChangeRoomReservationEvent(event, resources));
  };

  const onSelectSlot = (event) => {
    setSelectedEvent({
      title: '',
      start: moment(event.start).tz(utils.getUserTimezone()).format(),
      end: moment(event.end).tz(utils.getUserTimezone()).format(),
      resourceId: event.resourceId,
    });
    window.dispatchEvent(getAddRoomReservationEvent(event, resources));
  };

  /**
   * Handles a room reservation updated externally, such
   * as from a Drupal reservation form.
   *
   * @param {Object} values
   *  The incoming reservation values.
   * @param {String} values.id
   *  The reservation UUID.
   * @param {String} values.start
   *  The RFC 3339 formatted reservation start date time.
   * @param {String} values.end
   *  The RFC 3339 formatted reservation end date time.
   * @param {String} values.room
   *  The reservation's room UUID.
   */
  const onUpdateEvent = (values) => {
    const event = find(events, item => item.id === values.id) || {
      title: '',
    };
    // Apply the updated values to the stored event if found.
    setSelectedEvent({
      ...event,
      start: moment(values.start).tz(utils.getUserTimezone()).format(),
      end: moment(values.end).tz(utils.getUserTimezone()).format(),
      resourceId: values.room,
    });
  };

  const onSelectEvent = (event) => {
    setSelectedEvent(event);
    window.dispatchEvent(getSelectRoomReservationEvent(event.drupal_internal__id));
  };

  useEventListener(
    SAVE_ROOM_RESERVATION_SUCCESS,
    () => {
      doFetchReservations();
      // If we were creating a new reservation deselect the temp event.
      if (selectedEvent && typeof selectedEvent.id !== 'string') {
        setSelectedEvent(null);
      }
    },
    window,
    true
  );

  useEventListener(
    REFRESH_ROOM_RESERVATION,
    (event, values) => {
      onUpdateEvent(values);
    },
    window,
    true
  );

  useEventListener(CLOSE_ROOM_RESERVATION, () => {
    setSelectedEvent(null);
  });

  return (<div>
    <Calendar
      date={date}
      events={events}
      resources={resources}
      view={view}
      onDateChange={(values) => {
        purgeReservations();
        onChangeDate(values);
      }}
      onViewChange={onChangeView}
      onChangeEvent={onChangeEvent}
      onSelectEvent={onSelectEvent}
      onSelectSlot={onSelectSlot}
      onDoubleClickEvent={onDoubleClickEvent}
      selected={selectedEvent}
    />
  </div>);
};

RoomReservationScheduler.propTypes = {
  getLocationHours: PropTypes.func.isRequired,
  locations: PropTypes.arrayOf(Object).isRequired,
  locationsLoading: PropTypes.bool.isRequired,
  reservations: PropTypes.arrayOf(Object).isRequired,
  reservationsLoading: PropTypes.bool.isRequired,
  fetchLocations: PropTypes.func.isRequired,
  fetchReservations: PropTypes.func.isRequired,
  onChangeDate: PropTypes.func.isRequired,
  onChangeView: PropTypes.func.isRequired,
  purgeReservations: PropTypes.func.isRequired,
  date: PropTypes.instanceOf(Date),
  view: PropTypes.string,
};

RoomReservationScheduler.defaultProps = {
  view: 'day',
  date: utils.getUserTimeNow(),
};

const mapStateToProps = (state, ownProps) => ({
  locations: select.locationsAscending(state),
  locationsLoading: select.recordsAreLoading(c.TYPE_LOCATION)(state),
  getLocationHours: id => select.locationHoursTimesOnDate(id, ownProps.date || utils.getUserTimeNow())(state),
  reservations: select.roomReservationsArray(state),
  reservationsLoading: select.recordsAreLoading(c.TYPE_ROOM_RESERVATION)(state),
  openHoursLimit: select.locationsOpenHoursLimit(state),
  calendarRooms: [],
});

const mapDispatchToProps = dispatch => ({
  fetchLocations: (options) => {
    dispatch(api[c.TYPE_LOCATION].fetchAll(options));
  },
  fetchReservations: (options) => {
    dispatch(api[c.TYPE_ROOM_RESERVATION].fetchAll(options));
  },
  fetchUser: (options) => {
    dispatch(api[c.TYPE_USER].fetchAll(options));
  },
  purgeReservations: (options) => {
    dispatch(api[c.TYPE_ROOM_RESERVATION].purge(options));
  },
});

export default connect(
  mapStateToProps,
  mapDispatchToProps,
)(RoomReservationScheduler);
